@@ -43,13 +43,14 @@
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomevfs/gnome-vfs-volume-monitor.h>
 #include <libgnomevfs/gnome-vfs-drive.h>
-#include <glib/gi18n.h>
+#include <glib/gi18n-lib.h>
 
 #ifdef USE_GCONF
 #include <gconf/gconf-client.h>
 #endif
 
 #include <libgnomeui/gnome-icon-lookup.h>
+#include <libgnomeui/gnome-authentication-manager-private.h>
 
 #include <errno.h>
 #include <math.h>
@@ -81,6 +82,10 @@ struct _GtkFileSystemGnomeVFS
   GHashTable *folders;
 
   GnomeVFSVolumeMonitor *volume_monitor;
+  gulong volume_mounted_id;
+  gulong volume_unmounted_id;
+  gulong drive_connected_id;
+  gulong drive_disconnected_id;
 
 #ifdef USE_GCONF
   GConfClient *client;
@@ -120,7 +125,7 @@ struct _GtkFileFolderGnomeVFS
 
   GtkFileSystemGnomeVFS *system;
 
-  GHashTable *children;
+  GHashTable *children; /* NULL if destroyed */
 };
 
 typedef struct _FolderChild FolderChild;
@@ -129,8 +134,21 @@ struct _FolderChild
 {
   gchar *uri; /* canonical */
   GnomeVFSFileInfo *info;
+  guint reloaded : 1; /* used when reloading a folder */
 };
 
+/* When getting a folder, we launch two asynchronous processes, the monitor for
+ * changes and the async directory load.  Unfortunately, the monitor will feed
+ * us all the filenames.  So, to avoid duplicate emissions of the "files-added"
+ * signal, we keep a "reloaded" flag in each FolderChild structure.
+ *
+ * If we re-get a live folder, we don't re-create its monitor, but we do restart
+ * its async loader.  When we (re)read the information for a file, we turn on
+ * its "reloaded" flag.  When the async load terminates, we purge out all the
+ * structures which don't have the flag set, and emit "files-removed" signals
+ * for them.
+ */
+
 #ifdef USE_GCONF
 /* GConf paths for the bookmarks; keep these two in sync */
 #define BOOKMARKS_KEY  "/desktop/gnome/interface/bookmark_folders"
@@ -215,6 +233,7 @@ static void  gtk_file_folder_gnome_vfs_class_init (GtkFileFolderGnomeVFSClass *c
 static void  gtk_file_folder_gnome_vfs_iface_init (GtkFileFolderIface         *iface);
 static void  gtk_file_folder_gnome_vfs_init       (GtkFileFolderGnomeVFS      *impl);
 static void  gtk_file_folder_gnome_vfs_finalize   (GObject                    *object);
+static void  gtk_file_folder_gnome_vfs_dispose    (GObject                    *object);
 
 static GtkFileInfo *gtk_file_folder_gnome_vfs_get_info      (GtkFileFolder      *folder,
 							     const GtkFilePath  *path,
@@ -260,7 +279,8 @@ static void client_notify_cb (GConfClient *client,
 #endif
 
 static FolderChild *folder_child_new (const char       *uri,
-				      GnomeVFSFileInfo *info);
+				      GnomeVFSFileInfo *info,
+				      gboolean          reloaded);
 static void         folder_child_free (FolderChild *child);
 
 static void     set_vfs_error     (GnomeVFSResult   result,
@@ -278,7 +298,8 @@ static FolderChild *lookup_folder_child_from_uri (GtkFileFolder     *folder,
 static GObjectClass *system_parent_class;
 static GObjectClass *folder_parent_class;
 
-#define ITEMS_PER_NOTIFICATION 100
+#define ITEMS_PER_LOCAL_NOTIFICATION 10000
+#define ITEMS_PER_REMOTE_NOTIFICATION 100
 
 /*
  * GtkFileSystemGnomeVFS
@@ -382,15 +403,19 @@ gtk_file_system_gnome_vfs_init (GtkFileSystemGnomeVFS *system_vfs)
 							  NULL);
 #endif
 
-  system_vfs->volume_monitor = gnome_vfs_volume_monitor_ref (gnome_vfs_get_volume_monitor ());
-  g_signal_connect_object (system_vfs->volume_monitor, "volume-mounted",
-			   G_CALLBACK (volume_mount_unmount_cb), system_vfs, 0);
-  g_signal_connect_object (system_vfs->volume_monitor, "volume-unmounted",
-			   G_CALLBACK (volume_mount_unmount_cb), system_vfs, 0);
-  g_signal_connect_object (system_vfs->volume_monitor, "drive-connected",
-			   G_CALLBACK (drive_connect_disconnect_cb), system_vfs, 0);
-  g_signal_connect_object (system_vfs->volume_monitor, "drive-disconnected",
-			   G_CALLBACK (drive_connect_disconnect_cb), system_vfs, 0);
+  system_vfs->volume_monitor = gnome_vfs_get_volume_monitor ();
+  system_vfs->volume_mounted_id =
+    g_signal_connect_object (system_vfs->volume_monitor, "volume-mounted",
+			     G_CALLBACK (volume_mount_unmount_cb), system_vfs, 0);
+  system_vfs->volume_unmounted_id =
+    g_signal_connect_object (system_vfs->volume_monitor, "volume-unmounted",
+			     G_CALLBACK (volume_mount_unmount_cb), system_vfs, 0);
+  system_vfs->drive_connected_id =
+    g_signal_connect_object (system_vfs->volume_monitor, "drive-connected",
+			     G_CALLBACK (drive_connect_disconnect_cb), system_vfs, 0);
+  system_vfs->drive_disconnected_id =
+    g_signal_connect_object (system_vfs->volume_monitor, "drive-disconnected",
+			     G_CALLBACK (drive_connect_disconnect_cb), system_vfs, 0);
 
 }
 
@@ -409,7 +434,14 @@ gtk_file_system_gnome_vfs_finalize (GObject *object)
   gtk_file_paths_free (system_vfs->bookmarks);
 #endif
 
-  gnome_vfs_volume_monitor_unref (system_vfs->volume_monitor);
+  g_signal_handler_disconnect (system_vfs->volume_monitor, system_vfs->volume_mounted_id);
+  g_signal_handler_disconnect (system_vfs->volume_monitor, system_vfs->volume_unmounted_id);
+  g_signal_handler_disconnect (system_vfs->volume_monitor, system_vfs->drive_connected_id);
+  g_signal_handler_disconnect (system_vfs->volume_monitor, system_vfs->drive_disconnected_id);
+
+  /* FIXME: See http://bugzilla.gnome.org/show_bug.cgi?id=151463
+   * gnome_vfs_volume_monitor_unref (system_vfs->volume_monitor);
+   */
 
   /* XXX Assert ->folders should be empty
    */
@@ -518,27 +550,39 @@ gtk_file_system_gnome_vfs_get_volume_for_path (GtkFileSystem     *file_system,
 }
 
 
-static gboolean
-remove_all  (gpointer  key,
-	     gpointer  value,
-	     gpointer  user_data)
+static void
+unmark_all_fn  (gpointer  key,
+		gpointer  value,
+		gpointer  user_data)
 {
-  return TRUE;
+  FolderChild *child;
+
+  child = value;
+  child->reloaded = FALSE;
 }
 
 static void
 load_dir (GtkFileFolderGnomeVFS *folder_vfs)
 {
+  int num_items;
+
   if (folder_vfs->async_handle)
-    gnome_vfs_async_cancel (folder_vfs->async_handle);
+    {
+      gnome_vfs_async_cancel (folder_vfs->async_handle);
+      g_hash_table_foreach (folder_vfs->children, unmark_all_fn, NULL);
+    }
 
-  g_hash_table_foreach_remove (folder_vfs->children,
-			       remove_all, NULL);
   gnome_authentication_manager_push_async ();
+
+  if (g_str_has_prefix (folder_vfs->uri, "file:"))
+    num_items = ITEMS_PER_LOCAL_NOTIFICATION;
+  else
+    num_items = ITEMS_PER_REMOTE_NOTIFICATION;
+
   gnome_vfs_async_load_directory (&folder_vfs->async_handle,
 				  folder_vfs->uri,
 				  get_options (folder_vfs->types),
-				  ITEMS_PER_NOTIFICATION,
+				  num_items,
 				  GNOME_VFS_PRIORITY_DEFAULT,
 				  directory_load_callback, folder_vfs);
   gnome_authentication_manager_pop_async ();
@@ -775,6 +819,8 @@ volume_mount_cb (gboolean succeeded,
 {
   struct mount_closure *closure;
 
+  gdk_threads_enter ();
+  
   closure = data;
 
   closure->succeeded = succeeded;
@@ -786,6 +832,8 @@ volume_mount_cb (gboolean succeeded,
     }
 
   g_main_loop_quit (closure->loop);
+
+  gdk_threads_leave ();
 }
 
 static gboolean
@@ -805,9 +853,9 @@ gtk_file_system_gnome_vfs_volume_mount (GtkFileSystem        *file_system,
       gnome_vfs_drive_mount (GNOME_VFS_DRIVE (volume), volume_mount_cb, &closure);
       gnome_authentication_manager_pop_sync ();
 
-      GDK_THREADS_LEAVE ();
+      gdk_threads_leave ();
       g_main_loop_run (closure.loop);
-      GDK_THREADS_ENTER ();
+      gdk_threads_enter ();
       g_main_loop_unref (closure.loop);
 
       if (closure.succeeded)
@@ -1136,6 +1184,7 @@ gtk_file_system_gnome_vfs_parse (GtkFileSystem     *file_system,
       gchar *file;
       gchar *host_and_path;
       gchar *host_and_path_escaped;
+      gboolean complete_hostname = TRUE;
 
       gchar *colon = strchr (stripped, ':');
 
@@ -1148,6 +1197,9 @@ gtk_file_system_gnome_vfs_parse (GtkFileSystem     *file_system,
 
 	  host_name = g_strndup (colon + 1, first_slash - (colon + 1));
 
+	  if (first_slash == colon + 1)
+	    complete_hostname = FALSE;
+
 	  if (first_slash == last_slash)
 	    path = g_strdup ("/");
 	  else
@@ -1165,8 +1217,9 @@ gtk_file_system_gnome_vfs_parse (GtkFileSystem     *file_system,
 	  gchar *first_slash = strchr (colon + 3, '/');
 	  if (!first_slash)
 	    {
+	      complete_hostname = FALSE;
 	      host_name = g_strdup (colon + 3);
-	      path = g_strdup ("/");
+	      path = g_strdup ("");
 	      file = g_strdup ("");
 	    }
 	  else
@@ -1184,17 +1237,29 @@ gtk_file_system_gnome_vfs_parse (GtkFileSystem     *file_system,
 
       host_and_path = g_strconcat (host_name, path, NULL);
       host_and_path_escaped = gnome_vfs_escape_host_and_path_string (host_and_path);
-      *folder = gtk_file_path_new_steal (g_strconcat (scheme, "//", host_and_path_escaped, NULL));
-      *file_part = file;
-      result = TRUE;
+
+      if (complete_hostname)
+	{
+	  *folder = gtk_file_path_new_steal (g_strconcat (scheme, "//", host_and_path_escaped, NULL));
+	  *file_part = file;
+	  result = TRUE;
+	}
+      else
+	{
+	  /* Don't switch the folder until we have seen the full hostname.
+           * Otherwise, the auth dialog will come up for every single character
+           * of the hostname being typed in. 
+           */
+	  *folder = gtk_file_path_copy (base_path);
+	  *file_part = g_strdup (stripped);
+	  result = TRUE;
+	}
 
       g_free (scheme);
       g_free (host_name);
       g_free (path);
       g_free (host_and_path);
       g_free (host_and_path_escaped);
-
-      result = TRUE;
     }
   else
     {
@@ -1393,6 +1458,9 @@ gtk_file_system_gnome_vfs_render_icon (GtkFileSystem     *file_system,
   else
     pixbuf = NULL;
 
+  if (info)
+    gnome_vfs_file_info_unref (info);
+
   return pixbuf;
 }
 
@@ -1794,6 +1862,7 @@ gtk_file_folder_gnome_vfs_class_init (GtkFileFolderGnomeVFSClass *class)
   folder_parent_class = g_type_class_peek_parent (class);
 
   gobject_class->finalize = gtk_file_folder_gnome_vfs_finalize;
+  gobject_class->dispose = gtk_file_folder_gnome_vfs_dispose;
 }
 
 static void
@@ -1809,23 +1878,53 @@ gtk_file_folder_gnome_vfs_init (GtkFileFolderGnomeVFS *impl)
 {
 }
 
+static gboolean
+unref_at_idle (GObject *object)
+{
+  g_object_unref (object);
+  return FALSE;
+}
+
 static void
-gtk_file_folder_gnome_vfs_finalize (GObject *object)
+gtk_file_folder_gnome_vfs_dispose (GObject *object)
 {
   GtkFileFolderGnomeVFS *folder_vfs = GTK_FILE_FOLDER_GNOME_VFS (object);
   GtkFileSystemGnomeVFS *system_vfs = folder_vfs->system;
+  gboolean was_destroyed;
 
+  was_destroyed = folder_vfs->children == NULL;
+  
   if (folder_vfs->uri)
     g_hash_table_remove (system_vfs->folders, folder_vfs->uri);
+  folder_vfs->uri = NULL;
+  
   if (folder_vfs->async_handle)
     gnome_vfs_async_cancel (folder_vfs->async_handle);
+  folder_vfs->async_handle = NULL;
+  
   if (folder_vfs->monitor)
     gnome_vfs_monitor_cancel (folder_vfs->monitor);
+  folder_vfs->monitor = NULL;
+  
   if (folder_vfs->children)
     g_hash_table_destroy (folder_vfs->children);
+  folder_vfs->children = NULL;
 
-  g_free (folder_vfs->uri);
+  if (!was_destroyed)
+    {
+      /* The folder is now marked as destroyed, we
+       * free it at idle time to avoid races with the
+       * cancelled async job. See gnome_vfs_async_cancel()
+       */
+      g_object_ref (object);
+      g_idle_add ((GSourceFunc)unref_at_idle, object);
+    }
+}
 
+
+static void
+gtk_file_folder_gnome_vfs_finalize (GObject *object)
+{
   folder_parent_class->finalize (object);
 }
 
@@ -1862,7 +1961,7 @@ lookup_folder_child_from_uri (GtkFileFolder *folder,
     {
       GSList *uris;
 
-      child = folder_child_new (uri, vfs_info);
+      child = folder_child_new (uri, vfs_info, folder_vfs->async_handle ? TRUE : FALSE);
       gnome_vfs_file_info_unref (vfs_info);
       g_hash_table_replace (folder_vfs->children, child->uri, child);
 
@@ -1995,6 +2094,8 @@ info_from_vfs_info (const gchar      *uri,
 		    GtkFileInfoType   types)
 {
   GtkFileInfo *info = gtk_file_info_new ();
+  char *local_file;
+  char *display_name;
 
   if (types & GTK_FILE_INFO_DISPLAY_NAME)
     {
@@ -2007,9 +2108,16 @@ info_from_vfs_info (const gchar      *uri,
 	}
       else
 	{
-	  gchar *display_name = g_filename_to_utf8 (vfs_info->name, -1, NULL, NULL, NULL);
-	  if (!display_name)
-	    display_name = g_strescape (vfs_info->name, NULL);
+	  local_file = gnome_vfs_get_local_path_from_uri (uri);
+	  if (local_file != NULL)
+	    {
+	      display_name = g_filename_display_basename (local_file);
+	      g_free (local_file);
+	    }
+	  else
+	    {
+	      display_name = g_filename_display_name (vfs_info->name);
+	    }
 
 	  gtk_file_info_set_display_name (info, display_name);
 
@@ -2037,13 +2145,15 @@ info_from_vfs_info (const gchar      *uri,
 
 static FolderChild *
 folder_child_new (const char       *uri,
-		  GnomeVFSFileInfo *info)
+		  GnomeVFSFileInfo *info,
+		  gboolean          reloaded)
 {
   FolderChild *child;
 
   child = g_new (FolderChild, 1);
   child->uri = g_strdup (uri);
   child->info = info;
+  child->reloaded = reloaded ? TRUE : FALSE;
   gnome_vfs_file_info_ref (child->info);
 
   return child;
@@ -2111,7 +2221,9 @@ volume_mount_unmount_cb (GnomeVFSVolumeMonitor *monitor,
 			 GnomeVFSVolume        *volume,
 			 GtkFileSystemGnomeVFS *system_vfs)
 {
+  gdk_threads_enter ();
   g_signal_emit_by_name (system_vfs, "volumes-changed");
+  gdk_threads_leave ();
 }
 
 /* Callback used when a drive gets connected or disconnected */
@@ -2120,7 +2232,72 @@ drive_connect_disconnect_cb (GnomeVFSVolumeMonitor *monitor,
 			     GnomeVFSDrive         *drive,
 			     GtkFileSystemGnomeVFS *system_vfs)
 {
+  gdk_threads_enter ();
   g_signal_emit_by_name (system_vfs, "volumes-changed");
+  gdk_threads_leave ();
+}
+
+struct purge_closure
+{
+  GSList *removed_uris;
+};
+
+/* Used from g_hash_table_foreach() in folder_purge_and_unmark() */
+static gboolean
+purge_fn (gpointer key,
+	  gpointer value,
+	  gpointer data)
+{
+  struct purge_closure *closure;
+  FolderChild *child;
+
+  child = value;
+  closure = data;
+
+  if (child->reloaded)
+    {
+      child->reloaded = FALSE;
+      return FALSE;
+    }
+  else
+    {
+      closure->removed_uris = g_slist_prepend (closure->removed_uris, child->uri);
+
+      if (child->info)
+	gnome_vfs_file_info_unref (child->info);
+
+      g_free (child);
+
+      return TRUE;
+    }
+}
+
+/* Purges out all the files that don't have the "reloaded" flag set */
+static void
+folder_purge_and_unmark (GtkFileFolderGnomeVFS *folder_vfs)
+{
+  struct purge_closure closure;
+
+  closure.removed_uris = NULL;
+  g_hash_table_foreach_steal (folder_vfs->children, purge_fn, folder_vfs);
+
+  if (closure.removed_uris)
+    {
+      GSList *l;
+
+      closure.removed_uris = g_slist_reverse (closure.removed_uris);
+      g_signal_emit_by_name (folder_vfs, "files-removed", closure.removed_uris);
+      
+      for (l = closure.removed_uris; l; l = l->next)
+	{
+	  char *uri;
+
+	  uri = l->data;
+	  g_free (uri);
+	}
+
+      g_slist_free (closure.removed_uris);
+    }
 }
 
 static void
@@ -2135,11 +2312,20 @@ directory_load_callback (GnomeVFSAsyncHandle *handle,
   GSList *added_uris = NULL;
   GSList *changed_uris = NULL;
 
+  /* This is called from an idle, we need to protect is as such */
+  gdk_threads_enter ();
+
+  /* See if this folder is already destroyed.
+   * This can happen if we cancel in another thread while
+   * directory_load_callback is just being called
+   */
+  if (folder_vfs->children == NULL)
+    return;
+  
   for (tmp_list = list; tmp_list; tmp_list = tmp_list->next)
     {
       GnomeVFSFileInfo *vfs_info;
       gchar *uri;
-      gboolean new = FALSE;
 
       vfs_info = tmp_list->data;
       if (strcmp (vfs_info->name, ".") == 0 ||
@@ -2152,20 +2338,28 @@ directory_load_callback (GnomeVFSAsyncHandle *handle,
 	{
 	  FolderChild *child;
 
-	  child = folder_child_new (uri, vfs_info);
+	  child = g_hash_table_lookup (folder_vfs->children, uri);
+	  if (child)
+	    {
+	      child->reloaded = TRUE;
+
+	      if (child->info)
+		gnome_vfs_file_info_unref (child->info);
+
+	      child->info = vfs_info;
+	      gnome_vfs_file_info_ref (child->info);
+
+	      changed_uris = g_slist_prepend (changed_uris, child->uri);
+	    }
+	  else
+	    {
+	      child = folder_child_new (uri, vfs_info, TRUE);
+	      g_hash_table_insert (folder_vfs->children, child->uri, child);
+
+	      added_uris = g_slist_prepend (added_uris, child->uri);
+	    }
 
 	  g_free (uri);
-
-	  if (!g_hash_table_lookup (folder_vfs->children, child->uri))
-	    new = TRUE;
-
-	  g_hash_table_replace (folder_vfs->children,
-				child->uri, child);
-
-	  if (new)
-	    added_uris = g_slist_prepend (added_uris, child->uri);
-	  else
-	    changed_uris = g_slist_prepend (changed_uris, child->uri);
 	}
     }
 
@@ -2185,7 +2379,10 @@ directory_load_callback (GnomeVFSAsyncHandle *handle,
       folder_vfs->async_handle = NULL;
 
       g_signal_emit_by_name (folder_vfs, "finished-loading");
+      folder_purge_and_unmark (folder_vfs);
     }
+
+  gdk_threads_leave ();
 }
 
 static void
@@ -2198,6 +2395,13 @@ monitor_callback (GnomeVFSMonitorHandle   *handle,
   GtkFileFolderGnomeVFS *folder_vfs = user_data;
   GSList *uris;
 
+  /* This is called from an idle, we need to protect is as such */
+  gdk_threads_enter ();
+
+  /* Check if destroyed */
+  if (folder_vfs->children == NULL)
+    return;
+  
   switch (event_type)
     {
     case GNOME_VFS_MONITOR_EVENT_CHANGED:
@@ -2215,11 +2419,24 @@ monitor_callback (GnomeVFSMonitorHandle   *handle,
 	  {
 	    FolderChild *child;
 
-	    child = folder_child_new (info_uri, vfs_info);
-	    gnome_vfs_file_info_unref (vfs_info);
+	    child = g_hash_table_lookup (folder_vfs->children, info_uri);
 
-	    g_hash_table_replace (folder_vfs->children,
-				  child->uri, child);
+	    if (child)
+	      {
+		if (folder_vfs->async_handle)
+		  child->reloaded = TRUE;
+
+		if (child->info)
+		  gnome_vfs_file_info_unref (child->info);
+
+		child->info = vfs_info;
+	      }
+	    else
+	      {
+		child = folder_child_new (info_uri, vfs_info, folder_vfs->async_handle ? TRUE : FALSE);
+		gnome_vfs_file_info_unref (vfs_info);
+		g_hash_table_insert (folder_vfs->children, child->uri, child);
+	      }
 
 	    uris = g_slist_prepend (NULL, (gchar *)info_uri);
 	    if (event_type == GNOME_VFS_MONITOR_EVENT_CHANGED)
@@ -2244,6 +2461,8 @@ monitor_callback (GnomeVFSMonitorHandle   *handle,
     case GNOME_VFS_MONITOR_EVENT_METADATA_CHANGED:
       break;
     }
+  
+  gdk_threads_leave ();
 }
 
 static gboolean
